package proof

import proof.types.trees.BinaryTrees.{*, given}
import proof.types.formulas.*
import proof.types.formulas.ProofSystem.*

object InOrderMirrorProof extends InOrderMirrorTheorems:
  import InOrderMirror.*
  import Axioms.*
  import Variables.*
  import InductionHypotheses.*

  // Base Case
  val EmptyCase = Theorem(EmptyCaseStatement):
    Empty.mirror.inOrder
      === Empty.inOrder         ==< Mirror.EmptyCase
      === Nil                   ==< InOrder.EmptyCase
      === Nil.reverse           ==< Reverse.NilCase
      === Empty.inOrder.reverse ==< InOrder.EmptyCase

  // Inductive Case
  val NodeCase = Theorem(NodeCaseStatement):
    Node(left, x, right).inOrder.reverse
      // unfolding
      === (left.inOrder ++ ((x :: Nil) ++ right.inOrder)).reverse               ==< InOrder.NodeCase
      === ((x :: Nil) ++ right.inOrder).reverse ++ left.inOrder.reverse         ==< Reverse.AppendCase
      === (right.inOrder.reverse ++ (x :: Nil).reverse) ++ left.inOrder.reverse ==< Reverse.AppendCase
      === (right.inOrder.reverse ++ (x :: Nil)) ++ left.inOrder.reverse         ==< Reverse.SingleCase
      // folding
      === right.inOrder.reverse ++ ((x :: Nil) ++ left.inOrder.reverse)         ==< Append.Associativity
      === right.mirror.inOrder ++ ((x :: Nil) ++ left.inOrder.reverse)          ==< IHRight
      === right.mirror.inOrder ++ ((x :: Nil) ++ left.mirror.inOrder)           ==< IHLeft
      === Node(right.mirror, x, left.mirror).inOrder                            ==< InOrder.NodeCase
      === Node(left, x, right).mirror.inOrder                                   ==< Mirror.NodeCase

  @main def runTheorems =
    EmptyCase.verify
    NodeCase.verify
