package proof

import proof.types.trees.Lists.{*, given}
import proof.types.formulas.*
import proof.types.formulas.ProofSystem.*

/**
  * This file is a copy of exercise solutions, and is provided strictly as a
  * reference for proof syntax.
  */
object MapSingleFlatten:
  val f: T => List[T] = ListFunctionVariable("f")
  val x: T = Variable("x")
  val xs: List[T] = ListVariable("xs")
  val ys: List[T] = ListVariable("ys")
  val xss: List[List[T]] = ListVariable("xss")

  object Axioms:
    object Map:
      val NilCase   = Axiom(Nil.map(f) === Nil)
      val ConsCase  = Axiom((x :: xs).map(f) === f(x) :: xs.map(f))

    object Flatten:
      val NilCase   = Axiom(Nil.flatten === Nil)
      val ConsCase  = Axiom((xs :: xss).flatten === xs ++ xss.flatten)

    object Append:
      val NilCase   = Axiom(Nil ++ ys === ys)
      val ConsCase  = Axiom((x :: xs) ++ ys === x :: (xs ++ ys))

    object Single:
      val NilCase   = Axiom(single(x) === x :: Nil)
  end Axioms

  @main def examples =
    import Axioms.*
    // example of a theorem applying a single rule
    val exampleRuleApplication = Theorem(Nil.map(f) === Nil):
      Nil.map(f)
        === Nil ==< Map.NilCase

    // example of a theorem applying two rules successively
    val multipleRuleApplication = Theorem((Nil ++ Nil).map(f) === Nil):
      (Nil ++ Nil).map(f)
        === Nil.map(f)  ==< Append.NilCase
        === Nil         ==< Map.NilCase

    // a formula to use as an assumption
    val assumption = xs ++ Nil === Nil ++ xs

    // example of a theorem using an assumption
    // the assumption is automatically taken from the theorem statement
    val usingAssumption = Theorem(assumption ==> (xs ++ Nil === xs)):
      xs ++ Nil
        === Nil ++ xs ==< assumption // if assumption is not in the statement, this will fail
        === xs        ==< Append.NilCase

  @main def startingWithProofs =
    import Axioms.*

    val nilAppend = Theorem((x :: Nil) ++ Nil === (x :: Nil)):
      (x :: Nil) ++ Nil
        === x :: (Nil ++ Nil) ==< Append.ConsCase
        === x :: Nil          ==< Append.NilCase

    val assumption = xs ++ Nil === Nil ++ xs

    val concatNilFlatten = Theorem(
      assumption ==> ((xs ++ Nil.flatten) === xs)
    ):
      xs ++ Nil.flatten
        === xs ++ Nil ==< Flatten.NilCase
        === Nil ++ xs ==< assumption
        === xs        ==< Append.NilCase

  @main def mapSingle =
    import Axioms.*

    // Proof from the midterm exam
    //
    // MapSingleFlatten
    // Prove: forall xs. xs.map(single).flatten === xs
    //
    // As in the exam, the proof is in two parts:
    // 1. Prove the base case, list is Nil
    // 2. Prove the inductive case, list is x :: xs

    // Base Case
    val singleNil = Theorem(Nil.map(single).flatten === Nil):
      Nil.map(single).flatten
        === Nil.flatten ==< Map.NilCase
        === Nil         ==< Flatten.NilCase

    // Induction Hypothesis
    val IH = xs.map(single).flatten === xs

    // Inductive Case
    val singleInduct = Theorem(
      IH ==> ((x :: xs).map(single).flatten === (x :: xs))
    ):
      (x :: xs).map(single).flatten
        === (single(x) :: xs.map(single)).flatten   ==< Map.ConsCase
        === ((x :: Nil) :: xs.map(single)).flatten  ==< Single.NilCase
        === (x :: Nil) ++ xs.map(single).flatten    ==< Flatten.ConsCase
        === (x :: Nil) ++ xs                        ==< IH
        === x :: (Nil ++ xs)                        ==< Append.ConsCase
        === x :: xs                                 ==< Append.NilCase

end MapSingleFlatten
